/* dw - find duplicate words -- Standalone version in Standard C
**
** C/C++ standards conformance:
**	ANSI C  1989
**	ISO C   1990, 1995, 1999, 2011, 2018
**	ISO C++ 1998, 2003, 2011, 2014, 2017
**
** This program should be compilable by almost any C or C++ compiler,
** on any operating system, and on any CPU architecture.  It uses a
** few splint (static checker tool) annotations delimited by @
** characters, and embedded in comments, to disable bogus, or
** harmless, warning diagnostics.
**
** dw finds doubled words, normally ignoring lettercase, in standard
** input, or specified files.  Such words are reported with a line
** number, or number range, on standard output.
**
** Caveat: intervening punctuation hides doubled words, so filter
** input accordingly if that matters, such as with
**         sed -e 's/[.,:;/()<>\[\]{}]"/ /g' | dw
** or
**         tr -d '.,:;/()<>[]{}"' | dw
** or use the --skip option to treat punctuation like space.
**
** Usage:
**         dw [ --help ] [ --version ]
**     or
**         dw [ --keep-lettercase ] [ --list-words ] [ --skip-punctuation ]
**            [ --word-pattern '[initial][rest]' ] < infile [ > outfile ]
**     or
**         dw [ --keep-lettercase ] [ --list-words ] [ --skip-punctuation ]
**            [ --word-pattern '[initial][rest]' ]
**            infile1 infile2 infile3 ... [ > outfile ]
**
** The '[initial][rest]' argument can alternatively replaced by a pathname
** to a user-defined file, or the basename of an installed dw library file,
** that contains the pattern.  The library files are named with long and
** short ISO 639-1 alpha-2 codes for languages:
**
**	da.pat		danish.pat
**	de.pat		german.pat
**	es.pat		spanish.pat
**	fr.pat		french.pat
**	is.pat		icelandic.pat
**	...		...
**
** Thus, the option pairs -w fr and -w french choose the French
** pattern definition.
**
** A --debug option is also recognized, but it is primarily for
** developer use, and is intentionally not documented.
**
** [04-May-2019] -- version 2.00
*/

#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"

#ifdef __cplusplus
/*
** clang complains:
**     'register' storage class specifier is deprecated and incompatible with C++17
** so avoid its use for C++ compilation
*/
#define register
#endif

#if __STDC__
#define ARGS(plist)	plist
#else
#define ARGS(plist)	()
#endif

#ifndef PATH_MAX
#define PATH_MAX	4096		/* longest on current filesystem designs */
#endif

#if PATH_MAX < 1024			/* some systems define a rather small value */
#undef PATH_MAX
#define PATH_MAX	1024
#endif

#ifndef EXIT_FAILURE
#define EXIT_FAILURE	1
#endif

#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS	0
#endif

#define CHARSET_SIZE	256	/* assume 8-bit bytes: 2**8 == 256 */
#define CHARSET_MASK	0xff	/* mask to keep index of pat_xxx[] in range [0, CHARSET_SIZE - 1] */
#define DW_BUFSIZE	10240	/* was 1024 in dw-1.03 and earlier */

static char		last_word[DW_BUFSIZE];
static char		pat_initial[CHARSET_SIZE];
static char		pat_rest[CHARSET_SIZE];
static char		yytext[DW_BUFSIZE];
static const char	/*@observer@*/ *filename = "";
static const char	MATCH = (char)0xff;
static const char	*progbinpath = "";
static int		debug = 0;
static int		list_words = 0;
static int		lowercase = 1;
static int		skip_punctuation = 0;
static long		last_line_number = 0L;
static long		line_number = 1L;
static ptrdiff_t	yyleng = 0L;

static void		do_help		ARGS((void));
static void		do_one_file	ARGS((void));
static void		do_option_error	ARGS((const char * _option));
static void		do_version	ARGS((void));
static int		escape_char	ARGS((const char **_escape));

static const char *	fill_pattern	ARGS((char _pat_vector[], const char * _pattern));
static int		get_token	ARGS((void));
static int		hexval		ARGS((int _c));
static int		is_option	ARGS((const char * _expect, const char * _option));
static void		make_lower	ARGS((void));
static int		octval		ARGS((int _c));
static void		other		ARGS((void));
static void		word		ARGS((void));

int			main		ARGS((int _argc, char *_argv[]));

#define	T_EOF	EOF
#define	T_WORD	0
#define T_WHITE	1
#define T_OTHER	2

#if 0			/* old version with hardcoded word syntax */
#define	isinitial(c)	(isalpha(c) || ((c) == (int)'_'))
#define isrest(c)	(isinitial(c) || isdigit(c))
#else
#define CHAR_INDEX(c)	((int)(CHARSET_MASK & (int)(c)))
#define isinitial(c)	(pat_initial[CHAR_INDEX(c)] != '\0')
#define isrest(c)	(pat_rest [CHAR_INDEX(c)] != '\0')
#endif

#undef isodigit
#define isodigit(c)	(isdigit((int)(c)) && ((int)(c) != (int)'8') && ((int)(c) != (int)'9'))

#define FPRINTF	(void)fprintf
#define PRINTF	(void)printf


static void
do_help(void)
{
    PRINTF("\n");
    PRINTF("%s finds doubled words, normally ignoring lettercase, in standard\n", PACKAGE_NAME);
    PRINTF("input, or specified files.  Such words are reported with a line\n");
    PRINTF("number, or number range, on standard output.\n\n");
    PRINTF("Caveat: intervening punctuation hides doubled words, so filter\n");
    PRINTF("input accordingly if that matters, such as with\n");
    PRINTF("\tsed -e 's/[.,:;/()<>\\[\\]{}]\"/ /g' | %s\n", PACKAGE_NAME);
    PRINTF("or\n");
    PRINTF("\ttr -d '.,:;/()<>[]{}\"' | %s\n", PACKAGE_NAME);
    PRINTF("or use the --skip option to treat punctuation like space.\n\n");
    PRINTF("Usage:\n");
    PRINTF("\t%s [ --help ] [ --version ]\n", PACKAGE_NAME);
    PRINTF("    or\n");
    PRINTF("\t%s [ --keep-lettercase ] [ --list-words] [ --skip-punctuation ] [ --word-pattern '[initial][rest]' ] < infile [ > outfile ]\n", PACKAGE_NAME);
    PRINTF("    or\n");
    PRINTF("\t%s [ --keep-lettercase ] [ --list-words] [ --skip-punctuation ] [ --word-pattern '[initial][rest]' ] infile1 infile2 infile3 ... [ > outfile ]\n\n", PACKAGE_NAME);
    PRINTF("Options may be abbreviated to any unique prefix, with one or two leading\n");
    PRINTF("dashes, and their lettercase is NOT significant.\n\n");
}


static void
do_one_file(void)
{
    register int	token;

    while ((token = get_token()) != T_EOF)
    {
	switch (token)
	{
	case T_WORD:
	    word();
	    break;

	case T_WHITE:
	    break;

	case T_OTHER:
	    other();
	    break;

	default:
	    break;
	}
    }
}


static void
do_option_error(const char *option)
{
    PRINTF("%s: unrecognized option: [%s]\n", PACKAGE_NAME, option);
}


static void
do_version(void)
{
    PRINTF("%s %s\n", PACKAGE_NAME, PACKAGE_VERSION);
    PRINTF("Copyright (C) 1991-%s Free Software Foundation, Inc.\n", PACKAGE_YEAR);
    PRINTF("License GPLv2+: GNU GPL version 2 or later <https://gnu.org/licenses/gpl.html>.\n");
    PRINTF("This is free software; see the source for copying conditions.  There is NO\n");
    PRINTF("warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
}


static void
do_word_pattern(const char * pattern)
{
    const char *p;
    char pattern_file[PATH_MAX];
    char pattern_line[DW_BUFSIZE];

    while (isspace(pattern[0]))		/* ignore leading space */
	pattern++;

    if (pattern[0] != '[')		/* expect filename or pathname */
    {
	FILE *fpin;

	(void)strncpy(pattern_file, pattern, sizeof(pattern_file));
	fpin = fopen(pattern_file, "r");

	if (fpin == (FILE *)NULL)
	{	/* create something like "/usr/mysite/bin/../share/lib/dw/dw-2.00/fr.pat" */
	    char *q;

	    (void)strncpy(pattern_file, progbinpath, sizeof(pattern_file));
	    q = strrchr(pattern_file, '/');

	    if (q != (char *)NULL)	/* reduce "/usr/mysite/bin/dw" to "/usr/mysite/bin" */
		*q = '\0';

	    /* construct a pathname like "/usr/mysite/share/lib/dw/dw-2.00/fr.pat */
	    (void)strncat(pattern_file, "/../share/lib/",	sizeof(pattern_file) - 1);
	    (void)strncat(pattern_file, PACKAGE_TARNAME,	sizeof(pattern_file) - 1);
	    (void)strncat(pattern_file, "/",			sizeof(pattern_file) - 1);
	    (void)strncat(pattern_file, PACKAGE_TARNAME,	sizeof(pattern_file) - 1);
	    (void)strncat(pattern_file, "-",			sizeof(pattern_file) - 1);
	    (void)strncat(pattern_file, PACKAGE_VERSION,	sizeof(pattern_file) - 1);
	    (void)strncat(pattern_file, "/",			sizeof(pattern_file) - 1);
	    (void)strncat(pattern_file, pattern,		sizeof(pattern_file) - 1);
	    (void)strncat(pattern_file, ".pat",			sizeof(pattern_file) - 1);
	    pattern_file[sizeof(pattern_file) - 1] = '\0';
	    fpin = fopen(pattern_file, "r");
	}

	if (fpin == (FILE*)NULL)
	{
	    FPRINTF(stderr, "%s: cannot open pattern file [%s]\n",
		    PACKAGE_NAME, pattern_file);
	    exit(EXIT_FAILURE);
	}
	else
	{
	    char *s;

	    (void)memset(pattern_line, 0, sizeof(pattern_line));

	    do	/* skip sharp-initiated comment lines until first noncomment line */
	    {
	        s = fgets(pattern_line, (int)sizeof(pattern_line), fpin);
	    } while ( (s == pattern_line) && (pattern_line[0] == '#') );

	    if (s != pattern_line)
	    {
		FPRINTF(stderr, "%s: cannot read pattern file [%s]\n",
			PACKAGE_NAME, pattern_file);
		(void)fclose(fpin);
		exit(EXIT_FAILURE);
	    }
	    else
	    {
		(void)fclose(fpin);
		/*@-branchstate@*/
		pattern = (const char *)&pattern_line[0];
	    }
	}
    }

    (void)memset(pat_initial, 0, sizeof(pat_initial));
    (void)memset(pat_rest,    0, sizeof(pat_rest));
    /*@-mustfreefresh@*/
    p = fill_pattern(pat_initial, pattern);
    /*@-mustfreefresh@*/
    p = fill_pattern(pat_rest, p);

    if (debug != 0)
    {
	register int c;

	FPRINTF(stderr, "DEBUG: %s --word-pattern \"%s\"\n", PACKAGE_NAME, pattern);
	FPRINTF(stderr, "    dec\t    oct\t    hex\t   char\tinitial\t rest\n");

	for (c = 0; c < CHARSET_SIZE; ++c)
	{
	    FPRINTF(stderr, "    %3d\t    %03o\t     %02x\t      %c\t    %s\t    %s\n",
		    c,
		    (unsigned int)c,
		    (unsigned int)c,
		    (isprint(c) ? c : (int)'?'),
		    ((pat_initial[CHAR_INDEX(c)] == MATCH) ? "yes" : " no"),
		    ((pat_rest [CHAR_INDEX(c)] == MATCH) ? "yes" : " no"));
	}
    }

    if (p[0] != '\0')
    {
	FPRINTF(stderr, "%s: invalid word pattern = \"%s\"\n", PACKAGE_NAME, pattern);
	exit(EXIT_FAILURE);
    }
}


static int
escape_char(const char **escape)
{
    /*
    ** Expect *escape to be one of these:
    **     \ooo		# 3-digit octal value
    **     \oo		# 2-digit octal value
    **     \o		# 1-digit octal value
    **     \xhh		# 2-digit hexadecimal value
    **     \c		# literal single character c
    **     c		# any other single character
    **
    ** Return the value of the escaped character, and advance *escape to
    ** point to the last character parsed.
    **
    ** The returned value is guaranteed to be in the range [0, 0xff].
    */

    register int c;
    const unsigned char *p;

    p = (const unsigned char *)*escape;

    if (p[0] == (unsigned char)'\\')
    {
	if (p[1] == (unsigned char)'\0')
	    c = INT_MIN;
	else if ( ((p[1] == (unsigned char)'x') || (p[1] == (unsigned char)'X')) &&
		  isxdigit(p[2]) && 
		  isxdigit(p[3]))
	{				/* \xhh -> hexadecimal value */
	    /*@-shiftimplementation@*/
	    c = (hexval((int)p[2]) << 4) + hexval((int)p[3]);
	    p += 3;
	}
	else if (isodigit(p[1]) && isodigit(p[2]) && isodigit(p[3]))
	{				/* \ooo -> octal value */
	    /*@-shiftimplementation@*/
	    c = (octval((int)p[1]) << 6) + (octval((int)p[2]) << 3) + octval((int)p[3]);
	    p += 3;
	}
	else if (isodigit(p[1]) && isodigit(p[2]))
	{				/* \oo -> octal value */
	    /*@-shiftimplementation@*/
	    c = (octval((int)p[1]) << 3) + octval((int)p[2]);
	    p += 2;
	}
	else if (isodigit(p[1]))
	{				/* \o -> octal value */
	    c = octval((int)p[1]);
	    p++;
	}
	else				/* literal character */
	{
	    c = (int)p[1];
	    p++;
	}
    }
    else
	c = (int)p[0];

    assert ((0 <= c) && (c <= 0xff));

    return (c);
}


static const char *
fill_pattern(char pat_vector[], const char * pattern)
{		/* convert "[pattern]" to pat_vector[] entries */
    /*
    ** The pattern must be encased in brackets, with leading and
    ** trailing whitespace outside the brackets ignored.  The pattern
    ** consists of
    **
    **     ordinary characters		# other than [, ], \, or NUL
    **	   \ character			# literal character
    **     -				# range indication (as in A-Z)
    **
    ** To get [, ], \, and - into the pattern as ordinary characters,
    ** prefix them with backslash.
    */

    const char *p;
    int errors;
    int level;

    errors = 0;
    level = 0;

    while (isspace(*pattern))	/* skip leading whitespace */
	pattern++;

    p = pattern;
    (void)memset(pat_vector, 0, sizeof(pat_initial));

    if (p[0] == '[')
    {
	p++;
	level++;
    }
    else
	errors++;

    /*@-infloops@*/
    while ( (errors == 0) && (level > 0) )
    {
	register int c;

	c = escape_char(&p);

	switch (c)
	{
	case '\0':		/* premature end of pattern */
	    errors++;
	    break;

	case '[':
	    level++;

	    if (level > 1)
		pat_vector[CHAR_INDEX(c)] = MATCH;

	    break;

	case ']':
	    level--;

	    if (level > 0)
		pat_vector[CHAR_INDEX(c)] = MATCH;

	    break;

	default:		/* ordinary character, or start of range */
	    if (p[1] == '-')	/* character range */
	    {
		register int c1;
		register int c2;

		p += 2;
		c1 = c;
		c2 = escape_char(&p);

		if (c1 <= c2)
		{
		    for (c = c1; c <= c2; ++c)
			pat_vector[CHAR_INDEX(c)] = MATCH;

		    c = c2;	/* last explicit character parsed */
		}
		else
		    errors++;
	    }
	    else		/* ordinary character */
		pat_vector[CHAR_INDEX(c)] = MATCH;

	    break;
	}			/* end switch (p[0]) */

	p++;			/* advance past just-seen character */
    }			/* end while ( (errors == 0) && (level > 0) ) */

    while (isspace(p[0]))	/* allowing trailing whitespace */
	p++;

    if ( (p[0] != '\0') && (p[0] != '[') )
	errors++;

    if (errors > 0)
    {
	FPRINTF(stderr, "%s: invalid word pattern = \"%s\"\n", PACKAGE_NAME, pattern);
	exit(EXIT_FAILURE);
    }

    /*@-temptrans@*/
    return (p);
}


static int
get_token(void)
{
    register int	c;
    register char	*p;
    register int	token;

    p = yytext;
    c = getchar();

    if (c == EOF)
        token = T_EOF;
    else
    {
	size_t n;

	if (skip_punctuation != 0)	/* reduce consecutive punctuation to single space */
	{
	    /* NB: this test assumes isrest == (isinitial || isdigit):
	       that is OK only for current fixed word syntax */
	    while (ispunct(c) && !isrest(c))
		c = getchar();

	    (void)ungetc(c, stdin);	/* push back lookahead */
	    c = (int)' ';
	}

	if (isinitial(c))
	{
	    token = T_WORD;
	    n = 0;

	    while (isrest(c))
	    {
		n++;

		if (n < sizeof(yytext))
		    *p++ = (char)c;

		c = getchar();
	    }

	    (void)ungetc(c, stdin);	/* push back lookahead */
	}
	else if (isspace(c))	   	/* whitespace forms single token */
	{
	    token = T_WHITE;
	    n = 0;

	    while (isspace(c))
	    {
		if (c == (int)'\n')
		    line_number++;

		n++;

		if (n < sizeof(yytext))
		    *p++ = (char)c;

		c = getchar();
	    }

	    (void)ungetc(c, stdin);	/* push back lookahead */
	}
	else				/* all other tokens are single char */
	{
	    token = T_OTHER;
	}
    }

    *p = '\0';				/* terminate token in yytext[] */

    yyleng = p - yytext;		/* NB: no overflow possible because of type of yyleng */
					/* and limited size of yytext[] array */

    return (token);
}


static int
hexval(int c)
{
    int result;

    if (isupper(c))
	c = tolower(c);

    if (isdigit(c))
	result = c - (int)'0';
    else if (((int)'a' <= c) && (c <= (int)'f'))
	result = 10 + c - (int)'a';
    else
	result = INT_MIN;

    return(result);
}


static int
is_option(const char *expect, const char *option)
{
    /*
    ** Match option against expect, ignoring one or two leading hyphens,
    ** and lettercase, and accept the match if option is a prefix of
    ** at least one character of expect.  Return 1 for a match, and 0
    ** for a mismatch.
    */
    size_t count;

    if (expect[0] == '-') expect++;
    if (expect[0] == '-') expect++;

    if (option[0] == '-') option++;
    if (option[0] == '-') option++;

    count = 0;

    while ( (*expect != '\0') && (*option != '\0') && (tolower(*expect) == tolower(*option)) )
    {
	count++;
	option++;
	expect++;
    }

    return (((count > (size_t)0) && (*option == '\0')) ? 1: 0);
}


int
main(int argc, char *argv[])
{
    int errors;
    int n_command_line_files;
    int have_word_pattern;
    register int k;

    errors = 0;
    filename = "-";
    n_command_line_files = 0;
    have_word_pattern = 0;
    progbinpath = argv[0];

    if (argc > 1)
    {
	/*
	** In dw-1.03, we add support for two widely-implemented command-line
	** options in GNUware, and many non-GNU packages.  We permit option
	** to begin with one or two dashes, and to be abbreviated to unique
	** leading prefixes.  We also ignore lettercase.  Option response is on
	** stdout, following GNU coding standards.  Unrecognized
	** options terminate execution with EXIT_SUCCESS; unrecognized
	** ones terminate with EXIT_FAILURE.  Standard input is not
	** read if any --help or --version command-line option is
	** present.
	*/

	for (k = 1; k < argc; ++k)
	{
	    const char *option;

	    option = argv[k];

	    if (option[0] != '-') 	/* then argv[k] should be a pathname */
	    {
		n_command_line_files++;
		continue;
	    }

	    if (is_option("--debug", option) != 0) /* NB: intentionally undocumented! */
	    {
		debug = 1;
	    }
	    else if (is_option("--help", option) != 0)
	    {
		do_help();
		exit(EXIT_SUCCESS);
	    }
	    else if (is_option("--keep-lettercase", option) != 0)
	    {
		lowercase = 0;
	    }
	    else if (is_option("--list-words", option) != 0)
	    {
		list_words = 1;
	    }
	    else if (is_option("--skip-punctuation", option) != 0)
	    {
		skip_punctuation = 1;
	    }
	    else if (is_option("--version", option) != 0)
	    {
		do_version();
		exit(EXIT_SUCCESS);
	    }
	    else if (is_option("--word-pattern", option) != 0)
	    {
		if (k < (argc - 1))
		{
		    do_word_pattern(argv[++k]);
		    have_word_pattern++;
		    argv[k][0] = '\0';	/* hide file from further processing */
		}
		else
		{
		    FPRINTF(stderr, "%s: expected two bracketed patterns in one argument after --word-pattern\n",
			    PACKAGE_NAME);
		    exit(EXIT_FAILURE);
		}
	    }
	    else
	    {
		do_option_error(argv[k]);
		PRINTF("\n");
		do_help();
		exit(EXIT_FAILURE);
	    }
	}
    }

    if (have_word_pattern == 0)		/* fall back to traditional dw default of C-like identifier syntax */
	do_word_pattern("[A-Za-z_][A-Za-z_0-9]");

    last_word[0] = '\0';

    if (n_command_line_files == 0)	/* read stdin */
	do_one_file();
    else				/* read one or more command-line files */
    {
	FILE *fpin;

	fpin = (FILE *)NULL;

	for (k = 1; k < argc; ++k)
	{
	    if ( (argv[k][0] != '-') && (argv[k][0] != '\0') )
	    {
		filename = argv[k];
		fpin = freopen(filename, "r", stdin);

		if (fpin == (FILE *)NULL)
		{
		    errors++;
		    FPRINTF(stderr, "%s: ignoring unreadable file [%s]\n",
			    PACKAGE_NAME, argv[k]);
		}
		else
		    do_one_file();
	    }
	}

	if (fpin != (FILE *)NULL)
	    (void)fclose(fpin);
    }

    exit((errors == 0) ? EXIT_SUCCESS : EXIT_FAILURE);

    /*@notreached@*/
    /*@-warnlintcomments@*/
    /*NOTREACHED*/

    return (0);				/* keep optimizers happy */
}


static void
make_lower(void)
{
    if (lowercase != 0)
    {
	register ptrdiff_t n;

	for (n = 0; n < yyleng; ++n)
	{
	    if (isupper(yytext[n]))
		yytext[n] = (char)tolower(yytext[n]);
	}
    }
}


static int
octval(int c)
{
    int result;

    if (isodigit(c))
	result = c - (int)'0';
    else
	result = INT_MIN;

    return (result);
}


static void
other(void)
{
    assert(sizeof(last_word) >= sizeof(yytext));

    (void)strcpy(last_word, yytext);	/* so intervening 'words' do not */
					/* trigger output of duplicates */
}


static void
word(void)
{
    if (list_words != 0)
	(void)puts(yytext);

    make_lower();

    if (strcmp(yytext, last_word) == 0)
    {
	if (last_line_number == line_number)
	    PRINTF("%s:%ld: %s\n",
		   filename, line_number, yytext);
	else
	    PRINTF("%s:%ld-%ld: %s\n",
		   filename, last_line_number, line_number, yytext);
    }

    assert(sizeof(last_word) >= sizeof(yytext));

    (void)strcpy(last_word, yytext);
    last_line_number = line_number;
}
